import os
import platform
import shutil
import sys
from pathlib import Path
from typing import Iterator, Optional

from .runtime_spec import DotnetCoreRuntimeSpec


def find_dotnet_cli() -> Optional[Path]:
    dotnet_path = shutil.which("dotnet")
    if not dotnet_path:
        return None
    else:
        return Path(dotnet_path)


def find_dotnet_root() -> Path:
    """Try to discover the .NET Core root directory

    If the environment variable ``DOTNET_ROOT`` is defined, we will use that.
    Otherwise, we probe the default installation paths on Windows and macOS.

    If none of these lead to a result, we try to discover the ``dotnet`` CLI
    tool and use its (real) parent directory.

    Otherwise, this function raises an exception.

    :return: Path to the .NET Core root
    """

    dotnet_root = os.environ.get("DOTNET_ROOT", None)
    if dotnet_root is not None:
        return Path(dotnet_root)

    if sys.platform == "win32":
        # On Windows, the host library is stored separately from dotnet.exe for x86
        prog_files = os.environ.get("ProgramFiles")
        if not prog_files:
            raise RuntimeError("Could not find ProgramFiles")
        prog_files = Path(prog_files)
        dotnet_root = prog_files / "dotnet"
    elif sys.platform == "darwin":
        if "ARM64" in os.uname().version and platform.machine() == "x86_64":
            # Apple Silicon in Rosetta 2 mode
            dotnet_root = Path("/usr/local/share/dotnet/x64")
        else:
            dotnet_root = Path("/usr/local/share/dotnet")

    if dotnet_root is not None and dotnet_root.is_dir():
        return dotnet_root

    # Try to discover dotnet from PATH otherwise
    dotnet_cli = find_dotnet_cli()
    if not dotnet_cli:
        raise RuntimeError("Can not determine dotnet root")

    return dotnet_cli.resolve().parent


def find_runtimes_using_cli(dotnet_cli: Path) -> Iterator[DotnetCoreRuntimeSpec]:
    import re
    from subprocess import check_output

    out = check_output([str(dotnet_cli), "--list-runtimes"], encoding="UTF8")
    runtime_re = re.compile(r"(?P<name>\S+) (?P<version>\S+) \[(?P<path>[^\]]+)\]")

    for line in out.splitlines():
        m = re.match(runtime_re, line)
        if m:
            path = Path(m.group("path"))
            version = m.group("version")
            if path.is_dir():
                yield DotnetCoreRuntimeSpec(m.group("name"), version, path / version)


def find_runtimes_in_root(dotnet_root: Path) -> Iterator[DotnetCoreRuntimeSpec]:
    shared = dotnet_root / "shared"
    for runtime in shared.iterdir():
        if runtime.is_dir():
            name = runtime.name
            for version_path in runtime.iterdir():
                if version_path.is_dir():
                    yield DotnetCoreRuntimeSpec(name, version_path.name, version_path)


def find_runtimes() -> Iterator[DotnetCoreRuntimeSpec]:
    """Find installed .NET Core runtimes

    If the ``dotnet`` CLI can be found, we will call it as ``dotnet
    --list-runtimes`` and parse the result.

    If it is not available, we try to discover the dotnet root directory using
    :py:func:`find_dotnet_root` and enumerate the runtimes installed in the
    ``shared`` subdirectory.

    :return: Iterable of :py:class:`DotnetCoreRuntimeSpec` objects
    """
    dotnet_cli = find_dotnet_cli()
    if dotnet_cli is not None:
        return find_runtimes_using_cli(dotnet_cli)
    else:
        dotnet_root = find_dotnet_root()
        return find_runtimes_in_root(dotnet_root)


def find_libmono(*, sgen: bool = True) -> Path:
    """Find a suitable libmono dynamic library

    On Windows and macOS, we check the default installation directories.

    :param sgen:
        Whether to look for an SGen or Boehm GC instance. This parameter is
        ignored on Windows, as only ``sgen`` is installed with the default
        installer
    :return:
        Path to usable ``libmono``
    """
    unix_name = f"mono{'sgen' if sgen else ''}-2.0"
    if sys.platform == "win32":
        if sys.maxsize > 2**32:
            prog_files = os.environ.get("ProgramFiles")
        else:
            prog_files = os.environ.get("ProgramFiles(x86)")

        if prog_files is None:
            raise RuntimeError("Could not determine Program Files location")

        # Ignore sgen on Windows, the main installation only contains this DLL
        path = Path(prog_files) / "Mono/bin/mono-2.0-sgen.dll"

    elif sys.platform == "darwin":
        path = (
            Path("/Library/Frameworks/Mono.framework/Versions/Current/lib")
            / f"lib{unix_name}.dylib"
        )

    else:
        from ctypes.util import find_library

        path = find_library(unix_name)

    if path is None:
        raise RuntimeError("Could not find libmono")

    return Path(path)



from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Dict, Optional, Sequence

from .types import Assembly, Runtime, RuntimeInfo
from .util import StrOrPath
from .util.find import find_dotnet_root, find_libmono, find_runtimes
from .util.runtime_spec import DotnetCoreRuntimeSpec

__all__ = [
    "get_mono",
    "get_netfx",
    "get_coreclr",
    "find_dotnet_root",
    "find_libmono",
    "find_runtimes",
    "Runtime",
    "Assembly",
    "RuntimeInfo",
    "DotnetCoreRuntimeSpec",
]


def get_mono(
    *,
    # domain: Optional[str] = None,
    config_file: Optional[StrOrPath] = None,
    global_config_file: Optional[StrOrPath] = None,
    libmono: Optional[StrOrPath] = None,
    sgen: bool = True,
    debug: bool = False,
    jit_options: Optional[Sequence[str]] = None,
    assembly_dir: Optional[str] = None,
    config_dir: Optional[str] = None,
    set_signal_chaining: bool = False
) -> Runtime:
    """Get a Mono runtime instance

    :param config_file:
        Path to the domain configuration file
    :param global_config_file:
        Path to the global configuration file to load (defaults to, e.g.,
        ``/etc/mono/config``)
    :param libmono:
        Path to the Mono runtime dll/so/dylib. If this is not specified, we try
        to discover a globally installed instance using :py:func:`find_libmono`
    :param sgen:
        If ``libmono`` is not specified, this is passed to
        :py:func:`find_libmono`
    :param debug:
        Whether to initialise Mono debugging
    :param jit_options:
        "Command line options" passed to Mono's ``mono_jit_parse_options``
    :param assembly_dir:
        The base directory for assemblies, passed to ``mono_set_dirs``
    :param config_dir:
        The base directory for configuration files, passed to ``mono_set_dirs``
    :param set_signal_chaining:
        Whether to enable signal chaining, passed to ``mono_set_signal_chaining``.
        If it is enabled, the runtime saves the original signal handlers before
        installing its own, and calls the original ones in the following cases:
            - SIGSEGV/SIGABRT while executing native code
            - SIGPROF
            - SIGFPE
            - SIGQUIT
            - SIGUSR2
        This currently only works on POSIX platforms
    """
    from .mono import Mono

    libmono = _maybe_path(libmono)
    if libmono is None:
        libmono = find_libmono(sgen=sgen)

    impl = Mono(
        # domain=domain,
        debug=debug,
        jit_options=jit_options,
        config_file=_maybe_path(config_file),
        global_config_file=_maybe_path(global_config_file),
        libmono=libmono,
        assembly_dir=assembly_dir,
        config_dir=config_dir,
        set_signal_chaining=set_signal_chaining,
    )
    return impl


def get_coreclr(
    *,
    runtime_config: Optional[StrOrPath] = None,
    dotnet_root: Optional[StrOrPath] = None,
    properties: Optional[Dict[str, str]] = None,
    runtime_spec: Optional[DotnetCoreRuntimeSpec] = None,
) -> Runtime:
    """Get a CoreCLR (.NET Core) runtime instance

    The returned ``DotnetCoreRuntime`` also acts as a mapping of the config
    properties. They can be retrieved using the index operator and can be
    written until the runtime is initialized. The runtime is initialized when
    the first function object is retrieved.

    :param runtime_config:
        Pass to a ``runtimeconfig.json`` as generated by
        ``dotnet publish``. If this parameter is not given, a temporary runtime
        config will be generated.
    :param dotnet_root:
        The root directory of the .NET Core installation. If this is not
        specified, we try to discover it using :py:func:`find_dotnet_root`.
    :param properties:
        Additional runtime properties. These can also be passed using the
        ``configProperties`` section in the runtime config.
    :param runtime_spec:
        If the ``runtime_config`` is not specified, the concrete runtime to use
        can be controlled by passing this parameter. Possible values can be
        retrieved using :py:func:`find_runtimes`."""
    from .hostfxr import DotnetCoreRuntime

    dotnet_root = _maybe_path(dotnet_root)
    if dotnet_root is None:
        dotnet_root = find_dotnet_root()

    temp_dir = None
    runtime_config = _maybe_path(runtime_config)
    if runtime_config is None:
        if runtime_spec is None:
            candidates = [
                rt for rt in find_runtimes() if rt.name == "Microsoft.NETCore.App"
            ]
            candidates.sort(key=lambda spec: spec.version, reverse=True)
            if not candidates:
                raise RuntimeError("Failed to find a suitable runtime")

            runtime_spec = candidates[0]

        temp_dir = TemporaryDirectory()
        runtime_config = Path(temp_dir.name) / "runtimeconfig.json"

        with open(runtime_config, "w") as f:
            runtime_spec.write_config(f)

    impl = DotnetCoreRuntime(runtime_config=runtime_config, dotnet_root=dotnet_root)
    if properties:
        for key, value in properties.items():
            impl[key] = value

    if temp_dir:
        temp_dir.cleanup()

    return impl


def get_netfx(
    *, domain: Optional[str] = None, config_file: Optional[StrOrPath] = None
) -> Runtime:
    """Get a .NET Framework runtime instance

    :param domain:
        Name of the domain to create. If no value is passed, assemblies will be
        loaded into the root domain.
    :param config_file:
        Configuration file to use to initialize the ``AppDomain``. This will
        only be used for non-root-domains as we can not control the
        configuration of the implicitly loaded root domain.
    """
    from .netfx import NetFx

    impl = NetFx(domain=domain, config_file=_maybe_path(config_file))
    return impl


def _maybe_path(p: Optional[StrOrPath]) -> Optional[Path]:
    if p is None:
        return None
    else:
        return Path(p)



from pathlib import Path
from typing import Optional

from ..types import StrOrPath
from .clr_error import ClrError
from .coreclr_errors import get_coreclr_error
from .find import find_dotnet_root
from .hostfxr_errors import get_hostfxr_error

__all__ = [
    "check_result",
    "find_dotnet_root",
    "path_as_string",
    "optional_path_as_string",
]


def optional_path_as_string(path: Optional[StrOrPath]) -> Optional[str]:
    if path is None:
        return None
    return path_as_string(path)


def path_as_string(path: StrOrPath) -> str:
    return str(Path(path))


def check_result(err_code: int) -> None:
    """Check the error code of a .NET hosting API function and raise a
    converted exception.

    :raises ClrError: If the error code is `< 0`
    """

    if err_code < 0:
        hresult = err_code & 0xFFFF_FFFF
        error = get_coreclr_error(hresult)
        if not error:
            error = get_hostfxr_error(hresult)
        if not error:
            error = ClrError(hresult)
        raise error



#  tests for clr_loader-0.2.5-py39hca03da5_0 (this is a generated file);
print('===== testing package: clr_loader-0.2.5-py39hca03da5_0 =====');
print('running run_test.py');
#  --- run_test.py (begin) ---
try:
    from importlib.resources import files as resources_files
except ImportError:
    from importlib_resources import files as resources_files
import sys

import clr_loader


# Check if DLLs exist
base_path = resources_files('clr_loader') / 'ffi' / 'dlls'
for arch in ['amd64', 'x86']:
    for ext in ['dll', 'pdb']:
        if not (base_path / arch / f'ClrLoader.{ext}').exists():
            raise FileNotFoundError(f'DLL ffi/dlls/{arch}/ClrLoader.{ext} not found in package.')
print("DLLs found; test passed.")
#  --- run_test.py (end) ---

print('===== clr_loader-0.2.5-py39hca03da5_0 OK =====');



# flake8: noqa

cdef = [
    """
typedef void* pyclr_domain;
typedef int (*entry_point)(void* buffer, int size);

void pyclr_initialize();
void* pyclr_create_appdomain(const char* name, const char* config_file);
entry_point pyclr_get_function(pyclr_domain domain, const char* assembly_path, const char* class_name, const char* function);
void pyclr_close_appdomain(pyclr_domain domain);
void pyclr_finalize();
    """
]



# flake8: noqa

cdef = []

cdef.append(
    """
typedef struct _MonoDomain MonoDomain;
typedef struct _MonoAssembly MonoAssembly;
typedef struct _MonoImage MonoImage;
typedef struct _MonoMethodDesc MonoMethodDesc;
typedef struct _MonoMethod MonoMethod;
typedef struct _MonoObject MonoObject;

typedef enum {
	MONO_DEBUG_FORMAT_NONE,
	MONO_DEBUG_FORMAT_MONO,
	/* Deprecated, the mdb debugger is not longer supported. */
	MONO_DEBUG_FORMAT_DEBUGGER
} MonoDebugFormat;

char* mono_get_runtime_build_info (void);

MonoDomain* mono_jit_init(const char *root_domain_name);
void mono_jit_cleanup(MonoDomain *domain);
void mono_jit_parse_options(int argc, char * argv[]);

void mono_debug_init (MonoDebugFormat format);

MonoAssembly* mono_domain_assembly_open(MonoDomain *domain, const char *name);
MonoImage* mono_assembly_get_image(MonoAssembly *assembly);

void mono_domain_set_config(MonoDomain *domain, const char *base_dir, const char *config_file_name);
void mono_config_parse(const char* path);

MonoMethodDesc* mono_method_desc_new(const char* name, bool include_namespace);
MonoMethod* mono_method_desc_search_in_image(MonoMethodDesc *method_desc, MonoImage *image);
void mono_method_desc_free(MonoMethodDesc *method_desc);

MonoObject* mono_runtime_invoke(MonoMethod *method, void *obj, void **params, MonoObject **exc);

void* mono_object_unbox(MonoObject *object);

void mono_set_dirs(const char *assembly_dir, const char* config_dir);

void mono_set_signal_chaining(bool chain_signals);

"""
)



import sys
from pathlib import Path
from typing import Optional, Tuple

import cffi  # type: ignore

from . import hostfxr, mono, netfx

__all__ = ["ffi", "load_hostfxr", "load_mono", "load_netfx"]

ffi = cffi.FFI()  # type: ignore

for cdef in hostfxr.cdef + mono.cdef + netfx.cdef:
    ffi.cdef(cdef)


def load_hostfxr(dotnet_root: Path):
    hostfxr_name = _get_dll_name("hostfxr")
    dotnet_root = dotnet_root.absolute()

    # This will fail as soon as .NET hits version 10, but hopefully by then
    # we'll have a more robust way of finding the libhostfxr
    hostfxr_path = dotnet_root / "host" / "fxr"
    hostfxr_paths = hostfxr_path.glob(f"?.*/{hostfxr_name}")

    for hostfxr_path in reversed(sorted(hostfxr_paths, key=_path_to_version)):
        try:
            return ffi.dlopen(str(hostfxr_path))
        except Exception:
            pass

    try:
        return ffi.dlopen(str(dotnet_root / hostfxr_name))
    except Exception:
        pass

    raise RuntimeError(f"Could not find a suitable hostfxr library in {dotnet_root}")


def load_mono(path: Optional[Path] = None):
    # Preload C++ standard library, Mono needs that and doesn't properly link against it
    if sys.platform == "linux":
        ffi.dlopen("stdc++", ffi.RTLD_GLOBAL)

    path_str = str(path) if path else None
    return ffi.dlopen(path_str, ffi.RTLD_GLOBAL)


def load_netfx():
    if sys.platform != "win32":
        raise RuntimeError(".NET Framework is only supported on Windows")

    dirname = Path(__file__).parent / "dlls"
    if sys.maxsize > 2**32:
        arch = "amd64"
    else:
        arch = "x86"

    path = dirname / arch / "ClrLoader.dll"

    return ffi.dlopen(str(path))


def _path_to_version(path: Path) -> Tuple[int, int, int]:
    name = path.parent.name
    try:
        res = list(map(int, name.split(".")))
        return tuple(res + [0, 0, 0])[:3]
    except Exception:
        return (0, 0, 0)


def _get_dll_name(name: str) -> str:
    if sys.platform == "win32":
        return f"{name}.dll"
    elif sys.platform == "darwin":
        return f"lib{name}.dylib"
    else:
        return f"lib{name}.so"



import atexit
from pathlib import Path
from typing import Any, Optional

from .ffi import ffi, load_netfx
from .types import Runtime, RuntimeInfo, StrOrPath

_FW: Any = None


class NetFx(Runtime):
    def __init__(
        self, domain: Optional[str] = None, config_file: Optional[Path] = None
    ):
        initialize()
        if config_file is not None:
            config_file_s = str(config_file)
        else:
            config_file_s = ffi.NULL

        self._domain_name = domain
        self._config_file = config_file
        self._domain = _FW.pyclr_create_appdomain(domain or ffi.NULL, config_file_s)

    def info(self) -> RuntimeInfo:
        return RuntimeInfo(
            kind=".NET Framework",
            version="<undefined>",
            initialized=True,
            shutdown=_FW is None,
            properties=dict(
                domain=self._domain_name or "", config_file=str(self._config_file)
            ),
        )

    def _get_callable(self, assembly_path: StrOrPath, typename: str, function: str):
        func = _FW.pyclr_get_function(
            self._domain,
            str(Path(assembly_path)).encode("utf8"),
            typename.encode("utf8"),
            function.encode("utf8"),
        )

        return func

    def shutdown(self):
        if self._domain and _FW:
            _FW.pyclr_close_appdomain(self._domain)


def initialize():
    global _FW
    if _FW is not None:
        return

    _FW = load_netfx()
    _FW.pyclr_initialize()

    atexit.register(_release)


def _release():
    global _FW
    if _FW is not None:
        _FW.pyclr_finalize()
        _FW = None



# flake8: noqa

import sys


cdef = []

if sys.platform == "win32":
    cdef.append("typedef wchar_t char_t;")
else:
    cdef.append("typedef char char_t;")

# hostfxr.h
cdef.append(
    """
enum hostfxr_delegate_type
{
    hdt_com_activation,
    hdt_load_in_memory_assembly,
    hdt_winrt_activation,
    hdt_com_register,
    hdt_com_unregister,
    hdt_load_assembly_and_get_function_pointer
};

int32_t hostfxr_main(const int argc, const char_t **argv);

int32_t hostfxr_main_startupinfo(
    const int argc,
    const char_t **argv,
    const char_t *host_path,
    const char_t *dotnet_root,
    const char_t *app_path);

typedef void (*hostfxr_error_writer_fn)(const char_t *message);
hostfxr_error_writer_fn hostfxr_set_error_writer(hostfxr_error_writer_fn error_writer);

typedef void* hostfxr_handle;
typedef struct hostfxr_initialize_parameters
{
    size_t size;
    const char_t *host_path;
    const char_t *dotnet_root;
} hostfxr_initialize_parameters;

int32_t hostfxr_initialize_for_dotnet_command_line(
    int argc,
    const char_t **argv,
    const hostfxr_initialize_parameters *parameters,
    /*out*/ hostfxr_handle *host_context_handle);
int32_t hostfxr_initialize_for_runtime_config(
    const char_t *runtime_config_path,
    const hostfxr_initialize_parameters *parameters,
    /*out*/ hostfxr_handle *host_context_handle);

int32_t hostfxr_get_runtime_property_value(
    const hostfxr_handle host_context_handle,
    const char_t *name,
    /*out*/ const char_t **value);
int32_t hostfxr_set_runtime_property_value(
    const hostfxr_handle host_context_handle,
    const char_t *name,
    const char_t *value);
int32_t hostfxr_get_runtime_properties(
    const hostfxr_handle host_context_handle,
    /*inout*/ size_t * count,
    /*out*/ const char_t **keys,
    /*out*/ const char_t **values);

int32_t hostfxr_run_app(const hostfxr_handle host_context_handle);
int32_t hostfxr_get_runtime_delegate(
    const hostfxr_handle host_context_handle,
    enum hostfxr_delegate_type type,
    /*out*/ void **delegate);

int32_t hostfxr_close(const hostfxr_handle host_context_handle);
"""
)

# coreclr_delegates.h
cdef.append(
    """
// Signature of delegate returned by coreclr_delegate_type::load_assembly_and_get_function_pointer
typedef int (__stdcall *load_assembly_and_get_function_pointer_fn)(
    const char_t *assembly_path      /* Fully qualified path to assembly */,
    const char_t *type_name          /* Assembly qualified type name */,
    const char_t *method_name        /* Public static method name compatible with delegateType */,
    const char_t *delegate_type_name /* Assembly qualified delegate type name or null */,
    void         *reserved           /* Extensibility parameter (currently unused and must be 0) */,
    /*out*/ void **delegate          /* Pointer where to store the function pointer result */);

// Signature of delegate returned by load_assembly_and_get_function_pointer_fn when delegate_type_name == null (default)
typedef int (__stdcall *component_entry_point_fn)(void *arg, int32_t arg_size_in_bytes);
"""
)



import json
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Dict, TextIO


@dataclass
class DotnetCoreRuntimeSpec:
    """Specification of an installed .NET Core runtime"""

    name: str
    version: str
    path: Path

    @property
    def tfm(self) -> str:
        return f"net{self.version[:3]}"

    @property
    def floor_version(self) -> str:
        return f"{self.version[:3]}.0"

    @property
    def runtime_config(self) -> Dict[str, Any]:
        return {
            "runtimeOptions": {
                "tfm": self.tfm,
                "framework": {"name": self.name, "version": self.floor_version},
            }
        }

    def write_config(self, f: TextIO) -> None:
        json.dump(self.runtime_config, f)



from abc import ABCMeta, abstractmethod
from dataclasses import dataclass, field
from os import PathLike
from typing import Any, Callable, Dict, Optional, Union

__all__ = ["StrOrPath"]

StrOrPath = Union[str, PathLike]


@dataclass
class RuntimeInfo:
    """Information on a Runtime instance

    An informative text can be retrieved from this by converting it to a
    ``str``, in particular the following results in readable debug information:

        >>> ri = RuntimeInfo()
        >>> print(ri)
        6.12.0.122 (tarball)
        Runtime: Mono
        =============
          Version:      6.12.0.122 (tarball)
          Initialized:  True
          Shut down:    False
          Properties:
    """

    kind: str
    version: str
    initialized: bool
    shutdown: bool
    properties: Dict[str, str] = field(repr=False)

    def __str__(self) -> str:
        return (
            f"Runtime: {self.kind}\n"
            "=============\n"
            f"  Version:      {self.version}\n"
            f"  Initialized:  {self.initialized}\n"
            f"  Shut down:    {self.shutdown}\n"
            f"  Properties:\n"
            + "\n".join(
                f"    {key} = {_truncate(value, 65 - len(key))}"
                for key, value in self.properties.items()
            )
        )


class ClrFunction:
    def __init__(
        self, runtime: "Runtime", assembly: StrOrPath, typename: str, func_name: str
    ):
        self._assembly = assembly
        self._class = typename
        self._name = func_name

        self._callable = runtime._get_callable(assembly, typename, func_name)

    def __call__(self, buffer: bytes) -> int:
        from .ffi import ffi

        buf_arr = ffi.from_buffer("char[]", buffer)
        return self._callable(ffi.cast("void*", buf_arr), len(buf_arr))

    def __repr__(self) -> str:
        return f"<ClrFunction {self._class}.{self._name} in {self._assembly}>"


class Assembly:
    def __init__(self, runtime: "Runtime", path: StrOrPath):
        self._runtime = runtime
        self._path = path

    def get_function(self, name: str, func: Optional[str] = None) -> ClrFunction:
        """Get a wrapped .NET function instance

        The function must be ``static``, and it must have the signature
        ``int Func(IntPtr ptr, int size)``. The returned wrapped instance will
        take a ``binary`` and call the .NET function with a pointer to that
        buffer and the buffer length. The buffer is reflected using CFFI's
        `from_buffer`.

        :param name: If ``func`` is not given, this is the fully qualified name
                     of the function. If ``func`` is given, this is the fully
                     qualified name of the containing class
        :param func: Name of the function
        :return:     A function object that takes a single ``binary`` parameter
                     and returns an ``int``
        """
        if func is None:
            name, func = name.rsplit(".", 1)

        return ClrFunction(self._runtime, self._path, name, func)

    def __repr__(self) -> str:
        return f"<Assembly {self._path} in {self._runtime}>"


class Runtime(metaclass=ABCMeta):
    """CLR Runtime

    Encapsulates the lifetime of a CLR (.NET) runtime. If the instance is
    deleted, the runtime will be shut down.
    """

    @abstractmethod
    def info(self) -> RuntimeInfo:
        """Get configuration and version information"""
        pass

    def get_assembly(self, assembly_path: StrOrPath) -> Assembly:
        """Get an assembly wrapper

        This function does not guarantee that the respective assembly is or can
        be loaded. Due to the design of the different hosting APIs, loading only
        happens when the first function is referenced, and only then potential
        errors will be raised."""
        return Assembly(self, assembly_path)

    @abstractmethod
    def _get_callable(
        self, assembly_path: StrOrPath, typename: str, function: str
    ) -> Callable[[Any, int], Any]:
        """Private function to retrieve a low-level callable object"""
        pass

    @abstractmethod
    def shutdown(self) -> None:
        """Shut down the runtime as much as possible

        Implementations should still be able to "reinitialize", thus the final
        cleanup will usually happen in an ``atexit`` handler."""
        pass

    def __del__(self) -> None:
        self.shutdown()


def _truncate(string: str, length: int) -> str:
    if length <= 1:
        raise TypeError("length must be > 1")
    if len(string) > length - 1:
        return f"{string[:length-1]}…"
    else:
        return string
